@@ -6,13 +6,19 @@ from typing import Optional
 
 from django.conf import settings
 from django.contrib.auth import login
+from django.core.exceptions import PermissionDenied
 from django.db import models
 from django.http import HttpRequest
 from django.urls import reverse
 from django.utils import timezone
 
-from .exceptions import ExpiredToken, InactiveToken, InvalidTokenUse, UserMismatch
-from .settings import DEFAULT_EXPIRY, DEFAULT_REDIRECT
+from .exceptions import ExpiredLink, InactiveLink, InvalidLink, UsedLink
+from .settings import (
+    AUTHENTICATION_BACKEND,
+    DEFAULT_EXPIRY,
+    DEFAULT_REDIRECT,
+    SESSION_EXPIRY,
+)
 
 
 def parse_remote_addr(request: HttpRequest) -> str:
@@ -48,12 +54,18 @@ class MagicLink(models.Model):
         default=DEFAULT_REDIRECT,
     )
     created_at = models.DateTimeField(
-        default=timezone.now, help_text="When the token was originally created"
+        default=timezone.now, help_text="When the link was originally created"
     )
     expires_at = models.DateTimeField(
-        help_text="When the token is due to expire (uses DEFAULT_EXPIRY)",
+        help_text="When the link is due to expire (uses DEFAULT_EXPIRY)",
         default=link_expires_at,
     )
+    accessed_at = models.DateTimeField(
+        help_text="When the link was first used (GET)", blank=True, null=True
+    )
+    logged_in_at = models.DateTimeField(
+        help_text="When the link was used to login", blank=True, null=True
+    )
     is_active = models.BooleanField(
         default=True, help_text="Set to False to deactivate the token"
     )
@@ -65,63 +77,93 @@ class MagicLink(models.Model):
         return f"<MagicLink id={self.id} user_id={self.user_id} token='{self.token}'>'"
 
     def get_absolute_url(self) -> str:
-        return reverse("use_magic_link", kwargs={"token": self.token})
+        return reverse("magic_link", kwargs={"token": self.token})
 
     @property
     def has_expired(self) -> Optional[bool]:
-        """Return True if the token is past its expiry timestamp."""
+        """Return True if the link is past its expiry timestamp."""
         if self.expires_at:
             return self.expires_at < timezone.now()
         return None
 
+    @property
+    def has_been_used(self) -> bool:
+        """Return True if the link has been used to login already."""
+        return self.logged_in_at is not None
+
     @property
     def is_valid(self) -> bool:
-        """Return False if link has expired or been marked as inactive."""
-        return self.is_active and not self.has_expired
+        """Return True if the link can be used."""
+        return self.is_active and not self.has_expired and not self.has_been_used
 
-    def validate(self, request: HttpRequest) -> None:
+    def validate(self) -> None:
         """
-        Check token and request and raise InvalidToken if necessary.
+        Check token and request and raise InvalidLink if necessary.
 
-        In addition to the link itself, this method checks that the HttpRequest is
-        valid - which means that it is either unauthenticated, or authenticated as
-        the user to whom this link refers. You cannot use a magic link to log in as
-        someone else if you are already logged in.
+        This method checks the is_valid property, and if found to be False,
+        it then runs through the possibilities and raises an appropriate error.
 
         """
         if not self.is_active:
-            raise InactiveToken("Link is inactive")
+            raise InactiveLink("Link is inactive")
         if self.has_expired:
-            raise ExpiredToken("Link has expired")
-        if request.user.is_anonymous:
-            return
-        if request.user != self.user:
-            raise UserMismatch("User mismatch")
+            raise ExpiredLink("Link has expired")
+        if self.has_been_used:
+            raise UsedLink("Link has already been used")
+        # theoretically impossible, but belt-and-braces -
+        # ensures that is_valid and validate method are kept in sync.
+        if not self.is_valid:
+            raise InvalidLink("Link is invalid")
 
-    def log_use(
-        self, request: HttpRequest, error: InvalidTokenUse = None
+    def authorize(self, user: settings.AUTH_USER_MODEL) -> None:
+        """
+        Check the user may access this link.
+
+        Raises PermissionDenied if the user is authenticated already, and is not
+        the user defined in the link.
+
+        """
+        if user.is_authenticated and user != self.user:
+            raise PermissionDenied("User is already logged in as another user.")
+
+    def login(self, request: HttpRequest) -> None:
+        """Call login as the link.user."""
+        login(request, self.user, backend=AUTHENTICATION_BACKEND)
+        request.session.set_expiry(SESSION_EXPIRY)
+        self.logged_in_at = timezone.now()
+        self.save()
+
+    def disable(self) -> None:
+        """Disable the link regardless of expiry - used as a kill switch."""
+        self.is_active = False
+        self.save()
+
+    def audit(
+        self,
+        request: HttpRequest,
+        error: InvalidLink = None,
+        timestamp: datetime.datetime = None,
     ) -> MagicLinkUse:
-        """Create a MagicLinkUse from an HtttpRequest."""
-        return MagicLinkUse.objects.create(
+        """
+        Create a MagicLinkUse from an HtttpRequest.
+
+        The timestamp parameter is used to force the timestamp of the log to a specific
+        value - this is useful for aligning logs with parent link values.
+
+        """
+        log = MagicLinkUse.objects.create(
             link=self,
-            timestamp=timezone.now(),
+            timestamp=timestamp or timezone.now(),
             http_method=request.method,
             remote_addr=parse_remote_addr(request),
             ua_string=parse_ua_string(request),
             session_key=request.session.session_key or "",
-            link_is_valid=self.is_valid,
             error=str(error) if error else "",
         )
-
-    def login(self, request: HttpRequest) -> None:
-        """Call login as the link.user."""
-        login(request, self.user)
-
-    def disable(self) -> None:
-        """Disable the link (no further uses)."""
-        self.is_active = False
-        self.expires_at = timezone.now()
-        self.save()
+        if not self.accessed_at:
+            self.accessed_at = log.timestamp
+            self.save()
+        return log
 
 
 class MagicLinkUse(models.Model):
@@ -155,10 +197,6 @@ class MagicLinkUse(models.Model):
         help_text="The client User-Agent, extracted from HttpRequest headers",
         blank=True,
     )
-    link_is_valid = models.BooleanField(
-        help_text=("Snapshot of parent link is_valid property at the time of use"),
-        default=True,
-    )
     error = models.CharField(
         max_length=100,
         help_text="If the link use failed the error will be recorded here",
